import psycopg2

class PGDConnection:
    def __init__(self, connection_string: dict, query: str, values: tuple) -> None:
        self.query = query
        self.values = values
        self.connection = psycopg2.connect(
            host = connection_string['PG_HOST'],
            port = connection_string['PG_PORT'],
            database = connection_string['PG_DATABASE'],
            user = connection_string['PG_USER'],
            password = connection_string['PG_PASSWORD']
        )
        self.cursor = self.connection.cursor()

    def __enter__(self):
        self.cursor.execute(self.query, self.values)
        return self.cursor

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.connection.close()
        self.cursor.close


from typing import Awaitable
class PGDTransaction:
    @classmethod
    async def insert(cls, connection_object: any, main_table: str, column_and_value: dict, commit: bool = False) -> Awaitable[None]:
        """Dynamic query to insert certain table that can target several columns
        
        Params supplied for column_and_values are dict that contains key, value pair

        Usage: 
        example query -> INSERT INTO table(col_1, col_2) VALUES (val_1, val_2)
        input params ex-> column_and_value = {col_1: value_1, col_2: value_2}
        """
        try:
            column_keys = ', '.join([key for key in column_and_value.keys()])
            query_template = """
                INSERT INTO {main_table}({column_keys}) VALUES ({column_values})
            """
            query = query_template.format(main_table=main_table, column_keys=column_keys, column_values=', '.join(['%s']*(len(column_and_value))))
            value = tuple([value for value in column_and_value.values()])
            connection_object.cursor().execute(query, value)
            if commit:
                connection_object.commit()
            return
        except Exception as e:
            connection_object.rollback()
            raise e

    @classmethod
    async def update(cls, connection_object: any, main_table: str, column_and_value: dict, where: list, commit: bool = False) -> Awaitable[None]:
        """Dynamic query to update certain table that can takes multiple conditions and target several columns
        
        Params supplied for column_and_values and condition are dict that contains key, value pair
        Multiple conditions are concatenated in query using 'AND' clause
        """
        try:
            column_query = ', '.join([key + ' = %s' for key in column_and_value.keys()])
            where_query=''
            for w in where:
                where_query += f' {w["column_name"]} {w["operator"]} %s' if w.get('operator') else f' {w["column_name"]} = %s'
                where_query += f' {w["conjunction"]}' if w.get('conjunction') else ' '

            query_template = """
                UPDATE {main_table} SET {columns} WHERE {where_query}
            """
            query = query_template.format(main_table=main_table, columns=column_query, where_query=where_query)
            update_values = tuple(value for value in column_and_value.values())
            where_values = tuple(wv['value'] for wv in where)
            value = update_values + where_values
            connection_object.cursor().execute(query, value)
            if commit:
                connection_object.commit()
            return
        except Exception as e:
            connection_object.rollback()
            raise e

    @classmethod
    async def delete(cls, connection_object: any, main_table: str, where: list, commit: bool = False) -> Awaitable[None]:
        """Dynamic query to delete data from certain table that can takes several conditions
        
        Params supplied for conditions are dict that contains key, value pair
        Multiple conditions are concatenated in query using 'AND' clause
        """
        try:
            where_query=''
            for w in where:
                where_query += f' {w["column_name"]} {w["operator"]} %s' if w.get('operator') else f' {w["column_name"]} = %s'
                where_query += f' {w["conjunction"]}' if w.get('conjunction') else ' '

            query_template = """
                DELETE FROM {main_table} WHERE {where_query}
            """
            query = query_template.format(main_table=main_table, where_query=where_query)
            value = tuple(wv['value'] for wv in where)
            connection_object.cursor().execute(query, value)
            if commit:
                connection_object.commit()
            return
        except Exception as e:
            connection_object.rollback()
            raise e


from typing import Awaitable, Optional
from datetime import datetime, timedelta
from postgres_dynamic.pgd_connection import PGDConnection

class PGDGet:
    @classmethod
    async def get_one(cls, connection_string: dict, main_table: dict, where: list, join_table:list = [], column_name: list = None) -> Awaitable[dict]:
        try:
            result = {}
            column_name = ','.join(column_name) if column_name else '*'
            query = """
                SELECT {column_name} FROM {main_table} {main_table_alias} {join_table} WHERE {where_query}
            """
            
            where_query=''
            for w in where:
                where_query += f' {w["column_name"]} {w["operator"]} %s' if w.get('operator') else f' {w["column_name"]} = %s'
                where_query += f' {w["conjunction"]}' if w.get('conjunction') else ' '

            join_query = ''
            for i in join_table:
                join_query += ' INNER JOIN' if i['join_method'] == 'INNER' else ' LEFT JOIN' if i['join_method'] == 'LEFT' else ' RIGHT JOIN'
                join_query += f' {i["table"]} {i["alias"]} ON {i["on"]}'

            query = query.format(column_name=column_name, main_table=main_table['table'], main_table_alias=main_table['alias'] if main_table.get('alias') else '', join_table=join_query, where_query=where_query)
            where_values = tuple(wv['value'] for wv in where)
            values = where_values
            with PGDConnection(connection_string, query, values) as cursor:
                column_name = [i[0] for i in cursor.description]
                data = cursor.fetchone()
                if data:
                    for a, b in zip(column_name, data):
                        result[a] = str(b + timedelta(hours=7)) if type(b) is datetime else b
            return result
        except Exception as e:
            raise e

    @classmethod
    async def get_all(cls, connection_string: dict, main_table: dict, where: list = [], join_table:list = [], order: dict = {}, column_name: list = None, limit: Optional[int] = None, offset: Optional[int] = None) -> Awaitable[dict]:
        try:
            result = []
            column_name = ','.join(column_name) if column_name else '*'
            query = """
                SELECT {column_name} FROM {main_table} {main_table_alias} {join_table} {where_query} {order_query} LIMIT %s OFFSET %s
            """
            where_query = ''
            if where:
                where_query += 'WHERE '
                for w in where:
                    where_query += f' {w["column_name"]} {w["operator"]} %s' if w.get('operator') else f' {w["column_name"]} = %s'
                    where_query += f' {w["conjunction"]}' if w.get('conjunction') else ' '

            order_query = ''
            if order:
                order_query = 'ORDER BY ' 
                order_query += ''.join([f'{k} {v}' for k,v in order.items()]) if len(order) == 1 else ', '.join([f'{k} {v}' for k,v in order.items()])

            join_query = ''
            for i in join_table:
                join_query += ' INNER JOIN' if i['join_method'] == 'INNER' else ' LEFT JOIN' if i['join_method'] == 'LEFT' else ' RIGHT JOIN'
                join_query += f' {i["table"]} {i["alias"]} ON {i["on"]}'

            query = query.format(column_name=column_name, main_table=main_table['table'], main_table_alias=main_table['alias'] if main_table.get('alias') else '', join_table=join_query, where_query=where_query, order_query=order_query)
            where_values = tuple(wv['value'] for wv in where)
            offset_value = (limit*(offset-1)) if offset else None
            values = where_values + (limit, offset_value)
            with PGDConnection(connection_string, query, values) as cursor:
                column_name = [i[0] for i in cursor.description]
                for i in cursor.fetchall():
                    data = {}
                    for a, b in zip(column_name, i):
                        data[a] = str(b + timedelta(hours=7)) if type(b) is datetime else b
                    result.append(data)
            return {'data': result}
        except Exception as e:
            raise e

    @classmethod
    async def get_count(cls, connection_string: dict, main_table: dict, where: list = [], join_table:list = []) -> Awaitable[dict]:
        try:
            result = []
            query = """
                SELECT COUNT(*) FROM {main_table} {main_table_alias} {join_table} {where_query}
            """

            where_query = ''
            if where:
                where_query += 'WHERE '
                for w in where:
                    where_query += f' {w["column_name"]} {w["operator"]} %s' if w.get('operator') else f' {w["column_name"]} = %s'
                    where_query += f' {w["conjunction"]}' if w.get('conjunction') else ' '

            join_query = ''
            for i in join_table:
                join_query += ' INNER JOIN' if i['join_method'] == 'INNER' else ' LEFT JOIN' if i['join_method'] == 'LEFT' else ' RIGHT JOIN'
                join_query += f' {i["table"]} {i["alias"]} ON {i["on"]}'
                
            query = query.format(main_table=main_table['table'], main_table_alias=main_table['alias'] if main_table.get('alias') else '', join_table=join_query, where_query=where_query)
            where_values = tuple(wv['value'] if isinstance(wv['value'], str) else tuple(wv['value']) for wv in where )
            values = where_values
            with PGDConnection(connection_string, query, values) as cursor:
                result = cursor.fetchone()[0]
            return {'total_data': result}
        except Exception as e:
            raise e